#!/usr/bin/perl
use lib '/home/erealms/ethereal/mgmt/perl';

#################################################################################
# Created       : Martin Foster
# Modified      : 01-Jan-2007
#################################################################################
#
# Configure Ignore - Script part of Ethereal Realms, designed to manage 
#                    ignored user lists.
# Copyright (C) 2000-2007  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 4-3526 Wolfe Cres
#                 Halifax, Nova Scotia
#                 B3L 3S2
#
#################################################################################

use CGI qw(-no_debug -nosticky);				# Common gateway interface
use CGI::Carp qw(fatalsToBrowser);				# CGI Error logs
use strict;							# Strict variable enforcement

use Ethereal::Database;						# Database handler
use Ethereal::Login;						# Login functionality
use Ethereal::Option;						# Option management
use Ethereal::Menu;						# Consistent menu
use Ethereal::Param;						# Parameter control
use Ethereal::Template;						# Template handler

#################################################################################
# Gobal override
#################################################################################
$CGI::POST_MAX=1024 * 50;					# Maximum posts
$CGI::DISABLE_UPLOADS = 1;					# Disable uploads

#################################################################################
# Data Members
#################################################################################
my $database;							# Database handle
my $cgi;							# Common gateway interface handle
my $login;							# Login handle
my $menu;							# Menu system
my $param;							# Parameter Handle

my @args;							# Additional arguments

my %config;							# Configuration tag

#################################################################################
# Program Area
#################################################################################

	# Initial handles
	$cgi      = new CGI;
	$database = new Ethereal::Database();

	# Connect and fetch
	$database->Connect($cgi);

	# Set title
	$database->{'TITLE'} = $database->{'SYS'}{'TitIgnore'};


	# Parameter handling
	# Pre-fetch parameters
	@args = $cgi->param('REMOVE');

	# Fetch normally
	$param = new Ethereal::Param($database, $cgi);
	$param->GetParam();

	# Retreive hash
	$database->GetHashUser(\%config);
	

	# Create instances
	$menu  = new Ethereal::Menu($database, $param);
	$login = new Ethereal::Login($database, $cgi, $param, $menu);

	# Authenthication
	if ($login->GetVerificationNormal())
	{
		# Document header
		print $menu->Rewrite($database->DocumentGetHeader()), "\n";

		# Screen display
		Display($database, $cgi, $param, \%config, \@args);

		# Document footer
		print $database->DocumentGetFooter($database), "\n";
	}


#################################################################################
# Sub-Routines
#################################################################################

#####################
# Display
#
# Handles the general output of the initial display screen.  Which will allow one
# to change password, E-Mail addresses and/or passwords.

sub Display
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $config   = shift;					# Configuration hash
	my $args     = shift;					# Additional parameters

	my $option;						# Option handler
	my $table;						# Table handle
	my $tmpl;						# Template handle

	my $res;						# Results
	my $statement;						# Statement

	my $puppet;						# Choice of the matter	
	my $user;						# Username

	my $confirm;						# Confirmation fund
	my $ignored;						# List of the dammed
	my $types;						# Available ignore types

	my @types;						# Type of ignores
	my @ignored;						# List of the dammed
	

	#####################
	# Program area	
	
	# Create instances
	$option   = new Ethereal::Option();
	$tmpl     = new Ethereal::Template($config);


	# Information handling
	# Addition into list
	if (defined($param->{$config->{'TxtAddition'}}))
	{
		# Determine existence
		($puppet, $login) = $database->DataGet("SELECT
			  PuppetName,
			  PuppeteerLogin
			 FROM Puppet
			 WHERE PuppetName ILIKE ?", 

		 $param->{'IGNORE'}
		);


		# Discover
		$types = $option->FindIgnore($config->{'OptIgnore'}, $param->{'TYPE'});


		# Verification
		if (defined($puppet))
		{
			# Global or specific
			if ($types eq 'global')
			{
				# Block both puppet and puppeteer
				# Insert information
				$database->Quick("INSERT INTO PuppetIgnore 
					(PuppeteerLogin,
					 PuppetIgnore,
					 PuppetName,
					 PuppetLogin)
					VALUES (?,?,?,?)", 

			 	 $param->{'USER'}, 
				 $types,
				 $puppet,
				 $login
				);
			}
			else
			{
				# Simply a puppet
				# Insert information
				$database->Quick("INSERT INTO PuppetIgnore 
					(PuppeteerLogin,
					 PuppetIgnore,
					 PuppetName)
					VALUES (?,?,?)",

			 	 $param->{'USER'}, 
				 $types,
				 $puppet
				);
			}
		}
	}

	# Removal from list
	elsif (defined($param->{$config->{'TxtDelete'}}))
	{
		# Only if confirmed
		if (defined($param->{'CONFIRM'}))
		{
			# Loop and remove
			foreach $user (@args)
			{
				# Remove descriptor
				$user =~ s/^(\+|\-|\*|\~|\%)\s(.+)/$2/;

				# Remove ignored person
				$database->Quick("DELETE FROM PuppetIgnore 
					WHERE PuppeteerLogin=? 
					  AND PuppetName=?", 

				 $param->{'USER'}, 
				 $user
				);
			}
		}
	}

	# Cleanup
	$param->Cleanup('CONFIRM', 'GLOBAL');


	# Types
	$option->Split(\@types, $config->{'OptIgnore'});


	# Pretty adjustment
	$database->Pull(\$statement, "SELECT
		 PuppetName   AS \"PuppetName\",
		 PuppetIgnore AS \"PuppetIgnore\"
		FROM PuppetIgnore
		WHERE PuppeteerLogin=?
		ORDER BY PuppetName",

	 $param->{'USER'}
	);

	# Loop and create
	while ($res = $statement->fetchrow_hashref())
	{
		# Indicator
		# One
		if ($res->{'PuppetIgnore'} eq 'single')
		{
			# Normal application
			$puppet = "- $res->{PuppetName}";
		}

		# Many
		elsif ($res->{'PuppetIgnore'} eq 'global')
		{
			# Anti-social behaviour
			$puppet = "+ $res->{PuppetName}";
		}

		# Photographic
		elsif ($res->{'PuppetIgnore'} eq 'visual')
		{
			# Virgin eyes syndrome
			$puppet = "* $res->{PuppetName}";
		}

		# Beheading
		elsif ($res->{'PuppetIgnore'} eq 'behead')
		{
			# Virgin eyes syndrome
			$puppet = "% $res->{PuppetName}";
		}

		# Highlight
		else
		{
			# Virgin eyes syndrome
			$puppet = "~ $res->{PuppetName}";
		}

		# Push onto array
		push(@ignored, $puppet);

	}
	$statement->finish();

	
	# Generate widgets
	$confirm = $cgi->submit($config->{'TxtDelete'}) . '&nbsp;' . $cgi->checkbox(-name=>'CONFIRM', -label=>'');
	$ignored = $cgi->scrolling_list('REMOVE', \@ignored, $ignored[0], 10, 'true');
	$types   = $cgi->radio_group('TYPE', \@types, $types[0], 'true');


	# Informational display
	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";

	# Display template
	$tmpl->Show('TmplIgnore',
		WIGNORED => $ignored,
		WNTYPES  => $types,
		WNIGNORE => $cgi->textfield('IGNORE', '', 20, 30),
		WDELETE  => $confirm,
		WNSUBMIT => $cgi->submit($config->{'TxtAddition'})
	 );

	# From bottom
	print $cgi->end_form(), "\n";
}
